"""
Colorise v0.5.3 R21-R23

Author: Arttu Rautio (aturtur)
Website: http://aturtur.com/

#----------------------------------------------------------------------------------------------------------------------------------------
Change log:

v.0.5.3 (14.03.2021)
    > Cinema 4D R23 support

v.0.5.2
    > Added option to make polygon selections
    > Fixed reset color for non-Redshift nodes

v.0.5.1
    > Improved Redshift handling
    > Added option to colorise lights

"""
#----------------------------------------------------------------------------------------------------------------------------------------
# Libraries
#----------------------------------------------------------------------------------------------------------------------------------------
import os
import sys
import c4d
import maxon
import random as rnd

from c4d import storage as s
from c4d import gui, plugins, bitmaps, documents
from c4d.modules import colorchooser

try:
    import redshift
except:
    pass
#----------------------------------------------------------------------------------------------------------------------------------------
#  Register and initialize stuff
#----------------------------------------------------------------------------------------------------------------------------------------
PLUGIN_ID = 1041363 # Register plugin
#----------------------------------------------------------------------------------------------------------------------------------------
# Dictionaries and lists
#----------------------------------------------------------------------------------------------------------------------------------------
objectList = [ "BaseObject", "PointObject", "PolygonObject", "SplineObject", "InstanceObject",
              "HairObject", "VoronoiFracture", "CameraObject", "LodObject", "MotionTrackerObject",
              "FieldObject", "VolumeBuilder", "CAJointObject" ] # List of supported objects
colorsDict = { 11020: 11000, 11021: 11001, 11022: 11002, 11023: 11003, 11024: 11004, 11025: 11005,
              11026: 11006, 11027: 11007, 11028: 11008, 11029: 11009, 11030: 11010, 11031: 11011,
              11032: 11012, 11033: 11013, 11034: 11014, 11035: 11015 } # Link button to color field
colorFields = [ 11000, 11001, 11002, 11003, 11004, 11005, 11006, 11007, 11008, 11009, 11010, 11011, 11012, 11013, 11014, 11015 ]
materialSelections = [ 11020, 11021, 11022, 11023, 11024, 11025, 11026, 11027, 11028, 11029, 11030, 11031, 11032, 11033, 11034, 11035 ]
presets = {} # Dictionary to store all the preset paths
rsNodes = {
    # Materials
    "Architectural": c4d.Vector(0.529, 0.345, 0.333),         # RS Architectural
    "MatteShadow": c4d.Vector(0.529, 0.345, 0.333),           # RS Matte-Shadow Catcher
    "CarPaint": c4d.Vector(0.529, 0.345, 0.333),              # RS Car Paint
    "Hair": c4d.Vector(0.529, 0.345, 0.333),                  # RS Hair
    "Incandescent": c4d.Vector(0.529, 0.345, 0.333),          # RS Incandescent
    "Material": c4d.Vector(0.529, 0.345, 0.333),              # RS Material
    "MaterialBlender": c4d.Vector(0.529, 0.345, 0.333),       # RS Material Blender
    "Skin": c4d.Vector(0.529, 0.345, 0.333),                  # RS Skin
    "Sprite": c4d.Vector(0.529, 0.345, 0.333),                # RS Sprite
    "SubSurfaceScatter": c4d.Vector(0.529, 0.345, 0.333),     # RS SSS
    # Textures
    "NormalMap": c4d.Vector(0.663, 0.624, 0.424),             # RS Normal Map
    "AmbientOcclusion": c4d.Vector(0.663, 0.624, 0.424),      # RS AO
    "CameraMap": c4d.Vector(0.663, 0.624, 0.424),             # RS Camera Map
    "Curvature": c4d.Vector(0.663, 0.624, 0.424),             # RS Curvature
    "RSNoise": c4d.Vector(0.663, 0.624, 0.424),               # RS Noise
    "RSRamp": c4d.Vector(0.663, 0.624, 0.424),                # RS Ramp
    "TextureSampler": c4d.Vector(0.663, 0.624, 0.424),        # RS Texture
    "WireFrame": c4d.Vector(0.663, 0.624, 0.424),             # WireFrame
    # Utilities
    "StoreColorToAOV": c4d.Vector(0.69, 0.663, 0.78),         # RS Store Color To AOV
    "StoreIntegerToAOV": c4d.Vector(0.69, 0.663, 0.78),       # RS Store Integer To AOV
    "StoreScalarToAOV": c4d.Vector(0.69, 0.663, 0.78),        # RS Store Scalar To AOV
    "ParticleAttributeLookup": c4d.Vector(0.345, 0.31, 0.459),# RS Point Attribute
    "VertexAttributeLookup": c4d.Vector(0.345, 0.31, 0.459),  # RS Vertex Attribute
    "BumpBlender": c4d.Vector(0.345, 0.31, 0.459),            # RS Bump Blender
    "BumpMap": c4d.Vector(0.345, 0.31, 0.459),                # RS Bump Map
    "Displacement": c4d.Vector(0.345, 0.31, 0.459),           # RS Displacement
    "DisplacementBlender": c4d.Vector(0.345, 0.31, 0.459),    # RS Displacement Blender
    "C4DHairAttribute": c4d.Vector(0.533, 0.502, 0.639),      # RS C4D Hair Attributes
    "RSHairPosition": c4d.Vector(0.533, 0.502, 0.639),        # RS Hair Position
    "HairRandomColor": c4d.Vector(0.533, 0.502, 0.639),       # RS Hair Random Color
    "RaySwitch": c4d.Vector(0.533, 0.502, 0.639),             # RS Ray Switch
    "RSShaderSwitch": c4d.Vector(0.533, 0.502, 0.639),        # RS Shader Switch
    "RSUserDataColor": c4d.Vector(0.69, 0.663, 0.78),         # RS Color User Data
    "RSUserDataInteger": c4d.Vector(0.69, 0.663, 0.78),       # RS Integer User Data
    "RSUserDataScalar": c4d.Vector(0.69, 0.663, 0.78),        # RS Scalar User Data
    "RSUserDataVector": c4d.Vector(0.69, 0.663, 0.78),        # RS Vector User Data
    "Fresnel": c4d.Vector(0.424, 0.392, 0.541),               # RS Fresnel
    "RoundCorners": c4d.Vector(0.424, 0.392, 0.541),          # RS Round Corners
    "State": c4d.Vector(0.424, 0.392, 0.541),                 # RS State
    "TriPlanar": c4d.Vector(0.424, 0.392, 0.541),             # RS TriPlanar
    "RSVectorMaker": c4d.Vector(0.424, 0.392, 0.541),         # RS Vector Maker
    # Environment
    "Environment": c4d.Vector(0.761, 0.694, 0.784),           # RS Environment
    "Physical Sky": c4d.Vector(0.761, 0.694, 0.784),          # RS Physical Sky
    # Lights
    "Light_Dome": c4d.Vector(0.91, 0.8, 0.529),               # RS Dome Light
    "Light_IES": c4d.Vector(0.91, 0.8, 0.529),                # RS IES Light
    "Light": c4d.Vector(0.91, 0.8, 0.529),                    # RS Physical Light
    "Light_Portal": c4d.Vector(0.91, 0.8, 0.529),             # RS Portal Light
    "PhysicalSun": c4d.Vector(0.91, 0.8, 0.529),              # RS Sun Light
    # Volume
    "Volume": c4d.Vector(0.529, 0.494, 0.333),                # RS Volume
    # Math
    # Scalar
    "RSMathAbs": c4d.Vector(0.537, 0.71, 0.569),              # RS Abs
    "RSMathAdd": c4d.Vector(0.537, 0.71, 0.569),              # RS Add
    "RSMathACos": c4d.Vector(0.537, 0.71, 0.569),             # RS Arccosine
    "RSMathASin": c4d.Vector(0.537, 0.71, 0.569),             # RS Arcsine
    "RSMathATan2": c4d.Vector(0.537, 0.71, 0.569),            # RS ArcTan2
    "RSMathATan": c4d.Vector(0.537, 0.71, 0.569),             # RS Arctangent
    "RSMathBias": c4d.Vector(0.537, 0.71, 0.569),             # RS Bias
    "RSMathRange": c4d.Vector(0.537, 0.71, 0.569),            # RS Change Range
    "RSMathCos": c4d.Vector(0.537, 0.71, 0.569),              # RS Cosine
    "RSMathCrossVector": c4d.Vector(0.537, 0.71, 0.569),      # RS Cross Product
    "RSMathDiv": c4d.Vector(0.537, 0.71, 0.569),              # RS Div
    "RSMathDotVector": c4d.Vector(0.537, 0.71, 0.569),        # RS Dot Product
    "RSMathExp": c4d.Vector(0.537, 0.71, 0.569),              # RS Exp
    "RSMathFloor": c4d.Vector(0.537, 0.71, 0.569),            # RS Floor
    "RSMathFrac": c4d.Vector(0.537, 0.71, 0.569),             # RS Frac
    "RSMathGain": c4d.Vector(0.537, 0.71, 0.569),             # RS Gain
    "RSMathInv": c4d.Vector(0.537, 0.71, 0.569),              # RS Invert
    "RSMathLn": c4d.Vector(0.537, 0.71, 0.569),               # RS Ln
    "RSMathLog": c4d.Vector(0.537, 0.71, 0.569),              # RS Log
    "RSMathMax": c4d.Vector(0.537, 0.71, 0.569),              # RS Max
    "RSMathMin": c4d.Vector(0.537, 0.71, 0.569),              # RS Min
    "RSMathMix": c4d.Vector(0.537, 0.71, 0.569),              # RS Mix
    "RSMathMod": c4d.Vector(0.537, 0.71, 0.569),              # RS Mod
    "RSMathMul": c4d.Vector(0.537, 0.71, 0.569),              # RS Mul
    "RSMathNeg": c4d.Vector(0.537, 0.71, 0.569),              # RS Neg
    "RSMathNormalizeVector": c4d.Vector(0.537, 0.71, 0.569),  # RS Normalize
    "RSMathPow": c4d.Vector(0.537, 0.71, 0.569),              # RS Pow
    "RSMathRcp": c4d.Vector(0.537, 0.71, 0.569),              # RS Rcp
    "RSMathSaturate": c4d.Vector(0.537, 0.71, 0.569),         # RS Saturate
    "RSMathSign": c4d.Vector(0.537, 0.71, 0.569),             # RS Sign
    "RSMathSin": c4d.Vector(0.537, 0.71, 0.569),              # RS Sine
    "RSMathSqrt": c4d.Vector(0.537, 0.71, 0.569),             # RS Sqrt
    "RSMathSub": c4d.Vector(0.537, 0.71, 0.569),              # RS Sub
    "RSMathTan": c4d.Vector(0.537, 0.71, 0.569),              # RS Tangent
    # Vector
    "RSMathAbsVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Abs
    "RSMathAddVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Add
    "RSMathBiasVector": c4d.Vector(0.325, 0.51, 0.357),       # RS Vector Bias
    "RSMathRangeVector": c4d.Vector(0.325, 0.51, 0.357),      # RS Vector Change Range
    "RSMathDivVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Div
    "RSMathExpVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Exp
    "RSMathFloorVector": c4d.Vector(0.325, 0.51, 0.357),      # RS Vector Floor
    "RSMathFracVector": c4d.Vector(0.325, 0.51, 0.357),       # RS Vector Frac
    "RSMathGainVector": c4d.Vector(0.325, 0.51, 0.357),       # RS Vector Gain
    "RSMathInvVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Invert
    "RSMathLengthVector": c4d.Vector(0.325, 0.51, 0.357),     # RS Vector Length
    "RSMathLnVector": c4d.Vector(0.325, 0.51, 0.357),         # RS Vector Ln
    "RSMathLogVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Log
    "RSMathMaxVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Max
    "RSMathMinVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Min
    "RSMathMixVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Mix
    "RSMathModVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Mod
    "RSMathMulVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Mul
    "RSMathNegVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Neg
    "RSMathPowVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Pow
    "RSMathRcpVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Rcp
    "RSMathSaturateVector": c4d.Vector(0.325, 0.51, 0.357),   # RS Vector Saturate
    "RSMathSignVector": c4d.Vector(0.325, 0.51, 0.357),       # RS Vector Sign
    "RSMathSqrtVector": c4d.Vector(0.325, 0.51, 0.357),       # RS Vector Sqrt
    "RSMathSubVector": c4d.Vector(0.325, 0.51, 0.357),        # RS Vector Sub
    "RSVectorToScalars": c4d.Vector(0.325, 0.51, 0.357),      # RS Vector To Scalars
    # Color
    "RSMathAbsColor": c4d.Vector(0.788, 0.557, 0.537),        # RS Color Abs
    "RSMathBiasColor": c4d.Vector(0.788, 0.557, 0.537),       # RS Color Bias
    "RSColorRange": c4d.Vector(0.788, 0.557, 0.537),          # RS Color Change Range
    "RSColorComposite": c4d.Vector(0.788, 0.557, 0.537),      # RS Color Composite
    "RSColorConstant": c4d.Vector(0.788, 0.557, 0.537),       # RS Color Constant
    "RSColorCorrection": c4d.Vector(0.788, 0.557, 0.537),     # RS Color Correct
    "RSMathExpColor": c4d.Vector(0.788, 0.557, 0.537),        # RS Color Exp
    "RSMathGainColor": c4d.Vector(0.788, 0.557, 0.537),       # RS Color Gain
    "RSMathInvColor": c4d.Vector(0.788, 0.557, 0.537),        # RS Color Invert
    "RSColorLayer": c4d.Vector(0.788, 0.557, 0.537),          # RS Color Layer
    "RSColorMaker": c4d.Vector(0.788, 0.557, 0.537),          # RS Color Maker
    "RSColorMix": c4d.Vector(0.788, 0.557, 0.537),            # RS Color Mix
    "RSMathSaturateColor": c4d.Vector(0.788, 0.557, 0.537),   # RS Color Saturate
    "RSColorSplitter": c4d.Vector(0.788, 0.557, 0.537),       # RS Color Splitter
    "RSMathSubColor": c4d.Vector(0.788, 0.557, 0.537),        # RS Color Sub
    "RSColor2HSV": c4d.Vector(0.788, 0.557, 0.537),           # RS Color To HSV
    "RSHSV2Color": c4d.Vector(0.788, 0.557, 0.537),           # RS HSV To Color
    # Non-Redhsift
    1: c4d.Vector(97.0/255.0,98.0/255.0,100.0/255.0) # RS HSV To Color
} # Redshift nodes and their node colors
#----------------------------------------------------------------------------------------------------------------------------------------
# Functions
#----------------------------------------------------------------------------------------------------------------------------------------
def tool():
    """ Returns active tool data """
    doc = c4d.documents.GetActiveDocument()
    return doc.GetActiveToolData()

#----------------------------------------------------------------------------------------------------------------------------------------
def AssignToLayer(obj, color):
    """ Assigns object to layer """

    doc = c4d.documents.GetActiveDocument() # Get active document
    allLayers = CollectLayers() # Get all layers
    for layer in allLayers: # Iterate through layers
        if layer[c4d.ID_LAYER_COLOR] == color:  # If layer color is same as given color
            doc.AddUndo(c4d.UNDOTYPE_CHANGE, obj) # Record undo
            obj.SetLayerObject(layer) # Set layer
#----------------------------------------------------------------------------------------------------------------------------------------

def LoadSettings():
    """ Loads Colorise's settings from file and assign settings to global variables """

    global useIcons
    global useObjects
    global useLights
    global useXpresso
    global useDoodle
    global useRedshift
    global forceNulls
    global usePolygons

    try: # Try to ecexute following script
        path, fn = os.path.split(__file__)
        optionsFile = os.path.join(path, "res", "options.txt")

        if (sys.version_info >= (3, 0)): # If Python 3 version (R23)
            f = open(optionsFile) # Open the file for reading
        else: # If Python 2 version (R21)
            f = open(optionsFile.decode("utf-8"))

        optionsArray = [] # Initialize an array for options
        for line in f: # Iterate through every row
            line = line.rstrip('\n') # Strip newline stuff
            if line == 'True': # If line is true
                optionsArray.append(1) # Add 1 to the array
            else: # If line is false
                optionsArray.append(0) # Add 0 to the array
        f.close() # Close the file

        useIcons = optionsArray[0] # Colorise Icons
        useObjects = optionsArray[1] # Colorise Objects
        useLights = optionsArray[6] # Colorise Lights
        useXpresso = optionsArray[2] # Colorise Xpresso
        useDoodle = optionsArray[3] # Colorise Doodle Tool
        useRedshift = optionsArray[4] # Colorise Redshift
        forceNulls = optionsArray[5] # Force Nulls' Icon Color
        usePolygons = optionsArray[7] # Make polygon selections

    except: # If something went wrong
        pass # Do nothing
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def IdentifyObject(obj):
    """ Identifies object's type """

    objects = [ "PolygonObject", "SplineObject", "PointObject", "HairObject", "VolumeBuilder", "InstanceObject", "VoronoiFracture",
                "FieldObject", "CameraObject", "LodObject", "MotionTrackerObject", "CAJointObject" ]

    className = type(obj).__name__
    TypeName = obj.GetTypeName()

    if className == "XPressoTag":
        return "Xpresso"

    elif className == "BaseTag":
        return "Tag"

    elif className == "BaseObject":
        if TypeName == "Null":
            return "Null"
        return "Object"

    elif className in objects:
        return "Object"
#----------------------------------------------------------------------------------------------------------------------------------------
def GetNextItem(op):
    if op==None:
        return None
    if op.GetDown():
        return op.GetDown()
    while not op.GetNext() and op.GetUp():
        op = op.GetUp()
    return op.GetNext()


#----------------------------------------------------------------------------------------------------------------------------------------
# P O L Y G O N - S E L E C T I O N S
#----------------------------------------------------------------------------------------------------------------------------------------
def SelectionRemove(obj, name, selectedPolys):
    doc = c4d.documents.GetActiveDocument() # Get active document
    name = str(name) # Make sure name is a string
    tags = obj.GetTags() # Get tags
    for t in tags: # Iterate through tags
        if t.GetType() == 5673: # If tag is a selection tag
            if int(t.GetName()) in materialSelections: # If in names list
                if str(t.GetName()) != name: # But not this name
                    baseSelect = t.GetBaseSelect() # Get tag's base select
                    doc.AddUndo(c4d.UNDOTYPE_CHANGE, t) # Record undo
                    #baseSelect.Merge(selectedPolys) # Add to selection
                    #selectedPolys.CopyTo(baseSelect) # Replace selection
                    baseSelect.DeselectFrom(selectedPolys) # Remove from selection

def SelectionAdd(obj, selTag, selectedPolys):
    doc = c4d.documents.GetActiveDocument() # Get active document
    baseSelect = selTag.GetBaseSelect() # Get tag's base select
    doc.AddUndo(c4d.UNDOTYPE_CHANGE, selTag) # Record undo
    baseSelect.Merge(selectedPolys) # Add to selection
    #selectedPolys.CopyTo(baseSelect) # Replace selection

def CreateSelectionMaterial(obj, name, color):
    doc = c4d.documents.GetActiveDocument() # Get active document
    name = str(name) # Make sure name is a string
    mat = c4d.BaseMaterial(c4d.Mmaterial) # Initialize material
    mat.SetName(name) # Set name
    mat[c4d.MATERIAL_COLOR_COLOR] = color
    doc.InsertMaterial(mat) # Insert material to document
    doc.AddUndo(c4d.UNDOTYPE_NEW, mat) # Record undo for inserting a new material
    return mat

def CreateMaterialTag(obj, name):
    doc = c4d.documents.GetActiveDocument() # Get active document
    name = str(name) # Make sure name is a string
    matTag = c4d.BaseTag(5616) # Initialize a material tag
    matTag.SetName(name) # Set tag name
    obj.InsertTag(matTag) # Insert tag to object
    doc.AddUndo(c4d.UNDOTYPE_NEW, matTag) # Record undo for inserting a new tag
    return matTag

def CreateSelectionTag(obj, name, selectedPolys):
    doc = c4d.documents.GetActiveDocument() # Get active document
    name = str(name) # Make sure name is a string
    selTag = c4d.SelectionTag(c4d.Tpolygonselection) # Initialize a selection tag
    selTag.SetName(name) # Set tag name
    obj.InsertTag(selTag) # Insert tag to object
    doc.AddUndo(c4d.UNDOTYPE_NEW, selTag) # Record undo for inserting a new tag
    tagSelection = selTag.GetBaseSelect() # Initialize a base select
    selectedPolys.CopyTo(tagSelection) # Set selection
    return selTag

def GetCorrectTag(obj, name, tagType):
    name = str(name) # Make sure name is a string
    tags = obj.GetTags() # Get object's tags
    theTag = None # Initialize a variable
    for t in tags: # Iterate through tags
        if t.GetType() == tagType: # If tag is a selection tag
            if t.GetName() == name: # If correct selection tag
               theTag = t # Correct tag found
    if theTag != None: # If there is a correct tag
        return theTag # Return the tag
    else: # Otherwise
        return None # Return none
    
def GetCorrectMaterial(obj, name):
    doc = c4d.documents.GetActiveDocument() # Get active document
    name = str(name) # Make sure name is a string
    materials = doc.GetMaterials() # Get materials
    mat = None # Init state
    for m in materials: # Iterate through materials
        if name == m.GetName(): # If correct material found
            mat = m # Set mat
    return mat # Return material

def RemoveSelection(obj):
    doc = c4d.documents.GetActiveDocument() # Get active document
    if obj == None: return # Stop if no obj
    doc.StartUndo() # Start recording undos
    editorMode = doc.GetMode() # Get editor's active mode
    if editorMode == 7: # # If Polygons mode is activated
        selectedPolys = obj.GetPolygonS() # Get selected polygons
        if selectedPolys.GetCount() != 0: # If at least one polygon is selected
            tags = obj.GetTags() # Get tags
            for t in tags: # Iterate through tags
                if t.GetType() == 5673: # If tag is a selection tag
                    if int(t.GetName()) in materialSelections: # If in names list
                        baseSelect = t.GetBaseSelect() # Get tag's base select
                        doc.AddUndo(c4d.UNDOTYPE_CHANGE, t) # Record undo
                        baseSelect.DeselectFrom(selectedPolys) # Remove from selection

    #doc.ExecutePasses(None, 0, 1, 1, 0) # Needed when pressing buttons virtually
    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D

def CreateSelection(obj, name, color):
    doc = c4d.documents.GetActiveDocument() # Get active document
    if obj == None: return # Stop if no obj
    name = str(name) # Make sure name is a string
    doc.StartUndo() # Start recording undos
    editorMode = doc.GetMode() # Get editor's active mode
    if editorMode == 7: # # If Polygons mode is activated
        selectedPolys = obj.GetPolygonS() # Get selected polygons
        if selectedPolys.GetCount() != 0: # If at least one polygon is selected

            selTag = GetCorrectTag(obj, name, 5673) # Get correct selection tag
            matTag = GetCorrectTag(obj, name, 5616) # Get correct material tag
            matMat = GetCorrectMaterial(obj, name)  # Get correct material
            
            if (selTag == None and matTag == None and matMat == None): # Nothing found
                matMat = CreateSelectionMaterial(obj, name, color) # Create material
                selTag = CreateSelectionTag(obj, name, selectedPolys) # Create selection tag
                matTag = CreateMaterialTag(obj, name)# Create material tag
                matTag[c4d.TEXTURETAG_MATERIAL] = matMat # Set material
                matTag[c4d.TEXTURETAG_RESTRICTION] = selTag[c4d.ID_BASELIST_NAME] # Set selection tag
                SelectionRemove(obj, name, selectedPolys)
                
            elif (selTag != None and matTag == None and matMat == None): # Only selection tag found
                SelectionAdd(obj, selTag, selectedPolys) # Add to selection
                matMat = CreateSelectionMaterial(obj, name, color) # Create material
                matTag = CreateMaterialTag(obj, name)# Create material tag
                matTag[c4d.TEXTURETAG_MATERIAL] = matMat # Set material
                matTag[c4d.TEXTURETAG_RESTRICTION] = selTag[c4d.ID_BASELIST_NAME] # Set selection tag
                SelectionRemove(obj, name, selectedPolys)
                
            elif (selTag == None and matTag != None and matMat == None): # Only material tag found
                matMat = CreateSelectionMaterial(obj, name, color) # Create material
                selTag = CreateSelectionTag(obj, name, selectedPolys) # Create selection tag
                matTag[c4d.TEXTURETAG_MATERIAL] = matMat # Set material
                matTag[c4d.TEXTURETAG_RESTRICTION] = selTag[c4d.ID_BASELIST_NAME] # Set selection tag
                SelectionRemove(obj, name, selectedPolys)
                
            elif (selTag == None and matTag == None and matMat != None): # Only material found
                selTag = CreateSelectionTag(obj, name, selectedPolys) # Create selection tag
                matTag = CreateMaterialTag(obj, name)# Create material tag
                matTag[c4d.TEXTURETAG_MATERIAL] = matMat # Set material
                matTag[c4d.TEXTURETAG_RESTRICTION] = selTag[c4d.ID_BASELIST_NAME] # Set selection tag
                SelectionRemove(obj, name, selectedPolys)
                
            elif (selTag != None and matTag != None and matMat == None): # Selection tag and material tag found
                SelectionAdd(obj, selTag, selectedPolys) # Add to selection
                matMat = CreateSelectionMaterial(obj, name, color) # Create material
                matTag[c4d.TEXTURETAG_MATERIAL] = matMat # Set material
                matTag[c4d.TEXTURETAG_RESTRICTION] = selTag[c4d.ID_BASELIST_NAME] # Set selection tag
                SelectionRemove(obj, name, selectedPolys)
                
            elif (selTag != None and matTag == None and matMat != None): # Selection tag and material found
                SelectionAdd(obj, selTag, selectedPolys) # Add to selection
                matTag = CreateMaterialTag(obj, name)# Create material tag
                matTag[c4d.TEXTURETAG_MATERIAL] = matMat # Set material
                matTag[c4d.TEXTURETAG_RESTRICTION] = selTag[c4d.ID_BASELIST_NAME] # Set selection tag
                SelectionRemove(obj, name, selectedPolys)
                
            elif (selTag == None and matTag != None and matMat != None): # Material tag and material found
                selTag = CreateSelectionTag(obj, name, selectedPolys) # Create selection tag
                matTag[c4d.TEXTURETAG_MATERIAL] = matMat # Set material
                matTag[c4d.TEXTURETAG_RESTRICTION] = selTag[c4d.ID_BASELIST_NAME] # Set selection tag
                SelectionRemove(obj, name, selectedPolys)
                
            elif (selTag != None and matTag != None and matMat != None): # Everything found
                SelectionAdd(obj, selTag, selectedPolys) # Add to selection
                matTag[c4d.TEXTURETAG_MATERIAL] = matMat # Set material
                matTag[c4d.TEXTURETAG_RESTRICTION] = selTag[c4d.ID_BASELIST_NAME] # Set selection tag
                SelectionRemove(obj, name, selectedPolys)

    #doc.ExecutePasses(None, 0, 1, 1, 0) # Needed when pressing buttons virtually
    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D

def SelectPolygons(obj, addToSelection, id):
    """ Selects polygons that are colored by given color """
    doc = c4d.documents.GetActiveDocument() # Get active document
    if obj == None: return # Stop if no obj
    doc.StartUndo() # Start recording undos
    editorMode = doc.GetMode() # Get editor's active mode
    if editorMode == 7: # # If Polygons mode is activated
        baseSelect = obj.GetPolygonS() # Get selected polygons
        tags = obj.GetTags() # Get tags
        for t in tags: # Iterate through tags
            if t.GetType() == 5673:
                if t.GetName() == str(id):
                    if addToSelection == False:
                        baseSelect.DeselectAll()
                    baseSelect.Merge(t.GetBaseSelect())

def DeselectPolygons(obj, id):
    """ Deselects polygons that are colored by given color """
    doc = c4d.documents.GetActiveDocument() # Get active document
    if obj == None: return # Stop if no obj
    doc.StartUndo() # Start recording undos
    editorMode = doc.GetMode() # Get editor's active mode
    if editorMode == 7: # # If Polygons mode is activated
        baseSelect = obj.GetPolygonS() # Get selected polygons
        tags = obj.GetTags() # Get tags
        for t in tags: # Iterate through tags
            if t.GetType() == 5673:
                if t.GetName() == str(id):
                    baseSelect.DeselectFrom(t.GetBaseSelect())

#----------------------------------------------------------------------------------------------------------------------------------------
# O B J E C T S
#----------------------------------------------------------------------------------------------------------------------------------------
def SetColor(color, id):
    """ Sets given color to selected items """

    global useIcons
    global useObjects
    global useLights
    global useXpresso
    global useDoodle
    global useRedshift
    global forceNulls
    global usePolygons

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos
    selection = doc.GetSelection() # Get active selection

    for s in selection: # Iterate through selection
        
        # Colorise Xpresso
        if useXpresso == True: # If colorise Xpresso is true
            if IdentifyObject(s) == "Xpresso": # If Xpresso tag is found
                nm = s.GetNodeMaster() # Get node master
                root = nm.GetRoot() # Get node master root
                nm.AddUndo() # Record undo
                for node in root.GetChildren(): # Get Xpresso nodes
                    if node.GetBit(c4d.BIT_ACTIVE): # If node is selected
                        node[c4d.ID_GVBASE_COLOR] = color # Set node color
        
        # Colorise Objects
        if useObjects == True: # If colorise objects is true
            if type(s).__name__ in objectList: # If item's type is in object list
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
                s[c4d.ID_BASEOBJECT_USECOLOR] = 2 # Set 'Display Color' to 'On'
                s[c4d.ID_BASEOBJECT_COLOR] = color # Set object color
                if forceNulls == True:
                    if IdentifyObject(s) == "Null": # If object is a Null
                        s[c4d.ID_BASELIST_ICON_COLORIZE_MODE] = 1 # Custom
                        s[c4d.ID_BASELIST_ICON_COLOR] = color # Custom Icon Color
        
        # Colorise Icons
        if useIcons == True: # If colorise icons is true
            if type(s).__name__ in objectList: # If item's type is in object list
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
                s[c4d.ID_BASELIST_ICON_COLORIZE_MODE] = 1 # Custom
                s[c4d.ID_BASELIST_ICON_COLOR] = color # Custom Icon Color

        # Colorise Lights
        if useLights == True: # If colorise lights is true
            if s.GetType() == 5102: # If item is a light object
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
                s[c4d.LIGHT_COLOR] = color # Custom Light Color

        # Create polygon selections
        if usePolygons == True:
            objType = s.GetType() # Get object type
            if objType == 5100: # If object is a polygon object
                CreateSelection(s, id, color) # Create a selection

    # Colorise Redshift
    if useRedshift == True: # If colorise redshift is true
        materials = doc.GetMaterials() # Get materials
        try: # Try to execute following script
            for m in materials: # Iterate through materials
                if m.GetBit(c4d.BIT_ACTIVE): # If material is selected
                    rsnm = redshift.GetRSMaterialNodeMaster(m) # Get Redshift material node master
                    rsroot = rsnm.GetRoot() # Get node master root
                    rsnm.AddUndo() # Record undo
                    for c in rsroot.GetChildren(): # Iterate through nodes
                        if c.GetBit(c4d.BIT_ACTIVE): # If node is selected
                            c[c4d.ID_GVBASE_COLOR] = color # Set node color
        except: # Otherwise
            pass # Do nothing

    # Colorise Doodle Tool
    if useDoodle == True:
        #try: obj = doc.GetActiveObjects(False)[0] # If no object
        tool()[c4d.MDATA_DOODLE_COLOR] = color # Set doodle color

    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D

    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def SequenceColors(self):
    """ ... """

    global useIcons
    global useObjects
    global useLights
    global useXpresso
    global useDoodle
    global useRedshift
    global forceNulls
    global usePolygons

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos

    colors = [] # Initialize an array for colors
    for cf in colorFields: # Iterate through color fields
        colors.append(self.GetColorField(cf)["color"]) # Add color to an array

    #selection = doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_SELECTIONORDER) # Get selected items
    selection = doc.GetSelection() # Get selected items

    length = len(colors) # Get amount of the colors
    i = 0 # Initialize an iteration variable

    for s in selection: # Iterate through selection

        # Give colors to Xpresso nodes
        if useXpresso == True: # If colorise Xpresso is true
            if IdentifyObject(s) == "Xpresso": # If Xpresso tag is found
                nm = s.GetNodeMaster() # Get node master
                root = nm.GetRoot() # Get node master root
                nm.AddUndo() # Record undo
                k = 0 # Initialize an iteration variable
                color = c4d.Vector(0,0,0)
                for node in root.GetChildren(): # Get Xpresso nodes
                    if node.GetBit(c4d.BIT_ACTIVE): # If node is selected
                        node[c4d.ID_GVBASE_COLOR] = colors[k % length] # Set node color
                        k += 1 # One up!

        # Give colors to objects
        if useObjects == True:
            if type(s).__name__ in objectList: # If item's type is in object list
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record undo
                s[c4d.ID_BASEOBJECT_USECOLOR] = 2 # Set 'Display Color' to 'On'
                s[c4d.ID_BASEOBJECT_COLOR] = colors[i % length] # Set object color
                if forceNulls == True:
                    if IdentifyObject(s) == "Null": # If object is a Null
                        s[c4d.ID_BASELIST_ICON_COLORIZE_MODE] = 1 # Custom
                        s[c4d.ID_BASELIST_ICON_COLOR] = colors[i % length] # Custom Icon Color
        
        # Colorise Icons
        if useIcons == True: # If colorise icons is true
            if type(s).__name__ in objectList: # If item's type is in object list
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
                s[c4d.ID_BASELIST_ICON_COLORIZE_MODE] = 1 # Custom
                s[c4d.ID_BASELIST_ICON_COLOR] = colors[i % length] # Custom Icon Color

        # Colorise Lights
        if useLights == True: # If colorise lights is true
            if s.GetType() == 5102: # If item is a light object
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
                s[c4d.LIGHT_COLOR] = colors[i % length] # Custom Light Color

        i += 1 # Increase that little guy

    # Colorise Redshift
    if useRedshift == True: # If colorise redshift is true
        materials = doc.GetMaterials() # Get materials
        try: # Try to execute following script
            for m in materials: # Iterate through materials
                z = 0 # Bla bla bla
                if m.GetBit(c4d.BIT_ACTIVE): # If material is selected
                    rsnm = redshift.GetRSMaterialNodeMaster(m) # Get Redshift material node master
                    rsroot = rsnm.GetRoot() # Get node master root
                    rsnm.AddUndo() # Record undo
                    for c in rsroot.GetChildren(): # Iterate through nodes
                        if c.GetBit(c4d.BIT_ACTIVE): # If node is selected
                            c[c4d.ID_GVBASE_COLOR] = colors[z % length] # Set node color
                            z += 1 # One more time
        except: # Otherwise
            pass # Do nothing

    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def RandomColors(self):
    """ ... """

    global useIcons
    global useObjects
    global useLights
    global useXpresso
    global useDoodle
    global useRedshift
    global forceNulls
    global usePolygons

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos

    colors = [] # Initialize an array for colors
    for cf in colorFields: # Iterate through color fields
        colors.append(self.GetColorField(cf)["color"]) # Add color to an array

    selection = doc.GetSelection() # Get selected items

    for s in selection: # Iterate through selection

        # Give colors to Xpresso nodes
        if useXpresso == True: # If colorise Xpresso is true
            if IdentifyObject(s) == "Xpresso": # If Xpresso tag is found
                nm = s.GetNodeMaster() # Get node master
                root = nm.GetRoot() # Get node master root
                nm.AddUndo() # Record undo
                color = c4d.Vector(0,0,0)
                for node in root.GetChildren(): # Get Xpresso nodes
                    if node.GetBit(c4d.BIT_ACTIVE): # If node is selected
                        node[c4d.ID_GVBASE_COLOR] = rnd.choice(colors) # Set node color

        randomColor = rnd.choice(colors) # Get random color

        # Give colors to objects
        if useObjects == True:
            if type(s).__name__ in objectList: # If item's type is in object list
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record undo
                s[c4d.ID_BASEOBJECT_USECOLOR] = 2 # Set 'Display Color' to 'On'
                s[c4d.ID_BASEOBJECT_COLOR] = randomColor # Set object color
                if forceNulls == True:
                    if IdentifyObject(s) == "Null": # If object is a Null
                        s[c4d.ID_BASELIST_ICON_COLORIZE_MODE] = 1 # Custom
                        s[c4d.ID_BASELIST_ICON_COLOR] = randomColor # Custom Icon Color
        
        # Colorise Icons
        if useIcons == True: # If colorise icons is true
            if type(s).__name__ in objectList: # If item's type is in object list
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
                s[c4d.ID_BASELIST_ICON_COLORIZE_MODE] = 1 # Custom
                s[c4d.ID_BASELIST_ICON_COLOR] = randomColor # Custom Icon Color

        # Colorise Lights
        if useLights == True: # If colorise lights is true
            if s.GetType() == 5102: # If item is a light object
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
                s[c4d.LIGHT_COLOR] = randomColor # Custom Light Color

    # Colorise Redshift
    if useRedshift == True: # If colorise redshift is true
        materials = doc.GetMaterials() # Get materials
        try: # Try to execute following script
            for m in materials: # Iterate through materials
                if m.GetBit(c4d.BIT_ACTIVE): # If material is selected
                    rsnm = redshift.GetRSMaterialNodeMaster(m) # Get Redshift material node master
                    rsroot = rsnm.GetRoot() # Get node master root
                    rsnm.AddUndo() # Record undo
                    for c in rsroot.GetChildren(): # Iterate through nodes
                        if c.GetBit(c4d.BIT_ACTIVE): # If node is selected
                            c[c4d.ID_GVBASE_COLOR] = rnd.choice(colors) # Set node color
        except: # Otherwise
            pass # Do nothing

    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def ResetColor():
    """ Reset selected items display color """

    global useIcons
    global useObjects
    global useLights
    global useXpresso
    global useDoodle
    global useRedshift
    global forceNulls
    global usePolygons

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos
    selection = doc.GetSelection() # Get active selection

    for s in selection: # Iterate through selected items

        # Colorise Xpresso
        if useXpresso == True: # If colorise Xpresso is true
            if IdentifyObject(s) == "Xpresso": # If Xpresso tag is found
                nm = s.GetNodeMaster() # Get node master
                root = nm.GetRoot() # Get node master root
                nm.AddUndo() # Record undo
                for node in root.GetChildren(): # Get Xpresso nodes
                    if node.GetBit(c4d.BIT_ACTIVE): # If node is selected
                        node[c4d.ID_GVBASE_COLOR] = c4d.Vector(97.0/255.0,98.0/255.0,100.0/255.0) # Default Xpresso node color

        # Colorise Objects
        if useObjects == True: # If colorise objects is true
            if type(s).__name__ in objectList: # If item's type is in object list
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
                s[c4d.ID_BASEOBJECT_USECOLOR] = 0 # Set 'Display Color' to 'Off'
                s[c4d.ID_BASEOBJECT_COLOR] = doc[c4d.DOCUMENT_DEFAULTMATERIAL_COLOR] # Set color to documents default object color
                if forceNulls == True:
                    if IdentifyObject(s) == "Null": # If object is a Null
                        s[c4d.ID_BASELIST_ICON_COLORIZE_MODE] = 0 # Set 'Icon Color' to 'None'

        # Colorise Icons
        if useIcons == True: # If colorise icons is true
            doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
            s[c4d.ID_BASELIST_ICON_COLORIZE_MODE] = 0 # Set 'Icon Color' to 'None'

        # Colorise Lights
        if useLights == True: # If colorise lights is true
            if s.GetType() == 5102: # If item is a light object
                doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record change
                s[c4d.LIGHT_COLOR] = c4d.Vector(1,1,1) # Reset light color

        # Remove polygon selections
        if usePolygons == True: # If polygon stuff is enabled
            objType = s.GetType() # Get object type
            if objType == 5100: # If object is a polygon object
                RemoveSelection(s) # Remove selection



    # Colorise Redshift
    if useRedshift == True: # If colorise redshift is true
        materials = doc.GetMaterials() # Get materials
        try: # Try to execute following script
            for m in materials: # Iterate through materials
                if m.GetBit(c4d.BIT_ACTIVE): # If material is selected
                    rsnm = redshift.GetRSMaterialNodeMaster(m) # Get Redshift material node master
                    rsroot = rsnm.GetRoot() # Get node master root
                    rsnm.AddUndo() # Record undo
                    for c in rsroot.GetChildren(): # Iterate through nodes
                        if c.GetBit(c4d.BIT_ACTIVE): # If node is selected
                            opid = c.GetOperatorID() # Get operator ID
                            if opid == 1036227: # All the rest
                                if c[c4d.GV_REDSHIFT_SHADER_META_CLASSNAME] in rsNodes: # Check if node found in rsNodes dictionary
                                    c[c4d.ID_GVBASE_COLOR] = rsNodes[c[c4d.GV_REDSHIFT_SHADER_META_CLASSNAME]]
                                else:
                                    c[c4d.ID_GVBASE_COLOR] = c4d.Vector(97.0/255.0,98.0/255.0,100.0/255.0)
                            elif opid == 1036746: # Output
                                    c[c4d.ID_GVBASE_COLOR] = c4d.Vector(0.286, 0.329, 0.463)
                            elif opid == 1036762: # C4D Shader
                                c[c4d.ID_GVBASE_COLOR] = c4d.Vector(0.424, 0.392, 0.541)
                            elif opid == 1040213: # C4D Vertex Map
                                c[c4d.ID_GVBASE_COLOR] = c4d.Vector(0.424, 0.392, 0.541)
                            elif opid == 1040190: # RS MultiShader
                                c[c4d.ID_GVBASE_COLOR] = c4d.Vector(0.424, 0.392, 0.541)
                            elif opid == 1040192: # RS Reference
                                c[c4d.ID_GVBASE_COLOR] = c4d.Vector(0.424, 0.392, 0.541)
                            elif opid == 400001129: # Remark
                                c[c4d.ID_GVBASE_COLOR] = c4d.Vector(0.369, 0.431, 0.49)
                            elif opid == 400001120: # Constant
                                c[c4d.ID_GVBASE_COLOR] = c4d.Vector(0.537, 0.71, 0.569)
                            else:
                                c[c4d.ID_GVBASE_COLOR] = rsNodes[c[c4d.GV_REDSHIFT_SHADER_META_CLASSNAME]]
        except: # Otherwise
            pass # Do nothing

    # Colorise Doodle Tool
    if useDoodle == True:
        tool()[c4d.MDATA_DOODLE_COLOR] = c4d.Vector(1,1,1) # Set doodle color (white)

    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D
    return True # Everything is fine


#----------------------------------------------------------------------------------------------------------------------------------------
def SelectItem(color, addToSelection, id):
    """ Select items that are colored by given color """

    def SelectIfSameColor(op, color):
        """ Iterates through items and selects the item if it's color is same as given color """

        doc = c4d.documents.GetActiveDocument() # Get active document
        if op is None: # If no item
            return # Stop
        while op: # While there is item
            if op[c4d.ID_BASEOBJECT_COLOR] == color: # If display color is same as given color
                doc.AddUndo(c4d.UNDOTYPE_BITS, op) # Record undo
                op.SetBit(c4d.BIT_ACTIVE) # Select item
            op = GetNextItem(op) # Get next item
        return True

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos
    selection = doc.GetSelection() # Get active selection

    # Select Objects
    if useObjects == True:
        start_object = doc.GetFirstObject() # Get first object in the document
        if addToSelection == False:
            for s in selection: # Iterate through selected items
                doc.AddUndo(c4d.UNDOTYPE_BITS, s) # Record undo for deselecting item
                s.DelBit(c4d.BIT_ACTIVE) # Deselect the item
        SelectIfSameColor(start_object, color) # Iterate through object hierarchy and select if same color

    # Select Xpresso nodes
    if useXpresso == True:
        for s in selection: # Iterate through selected items
            if IdentifyObject(s) == "Xpresso": # If item is Xpresso
                nm = s.GetNodeMaster() # Get node master
                root = nm.GetRoot() # Get node master's root
                nm.AddUndo() # Record undo
                if addToSelection == False:
                    for c in root.GetChildren(): # Iterate through nodes
                        c.DelBit(c4d.BIT_ACTIVE) # Deselect node
                for c in root.GetChildren(): # Iterate through nodes
                    if c[c4d.ID_GVBASE_COLOR] == color: # If node color is same as given color
                        c.SetBit(c4d.BIT_ACTIVE) # Select node

    # Select Redshift nodes
    if useRedshift == True:
        materials = doc.GetMaterials() # Get materials
        try: # Try to execute following script
            for m in materials: # Iterate through materials
                if m.GetBit(c4d.BIT_ACTIVE): # If material is selected
                    rsnm = redshift.GetRSMaterialNodeMaster(m) # Get Redshift material node master
                    rsroot = rsnm.GetRoot() # Get node master's root
                    rsnm.AddUndo() # Record undo
                    if addToSelection == False:
                        for r in rsroot.GetChildren(): # Iterate through nodes
                            r.DelBit(c4d.BIT_ACTIVE) # Deselect node
                    for r in rsroot.GetChildren(): # Iterate through nodes
                        if r[c4d.ID_GVBASE_COLOR] == color: # If node color is same as given color
                            r.SetBit(c4d.BIT_ACTIVE) # Select node
        except: # If something went wrong
            pass # Do nothing


    # Select polygons
    if usePolygons == True:
        for s in selection: # Iterate through selected objects
            if s.GetType() == 5100: # If object is a polygon object
                SelectPolygons(s, addToSelection, id)

    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D
    return True # Everything is fine
#----------------------------------------------------------------------------------------------------------------------------------------
def RemoveFromSelection(color, id):
    """ Removes items from selection that are colored by given color """

    def DeselectIfSameColor(op, color):
        """ Iterates through items and deselects the item if it's color is same as given color """

        doc = c4d.documents.GetActiveDocument() # Get active document
        if op is None: # If no item
            return # Stop
        while op: # While there is item
            if op[c4d.ID_BASEOBJECT_COLOR] == color: # If display color is same as given color
                doc.AddUndo(c4d.UNDOTYPE_BITS, op) # Record undo
                op.DelBit(c4d.BIT_ACTIVE) # Deselect item
            op = GetNextItem(op) # Get next item
        return True

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos
    selection = doc.GetSelection() # Get active selection

    # Select Objects
    if useObjects == True:
        start_object = doc.GetFirstObject() # Get first object in the document
        for s in selection: # Iterate through selected items
            DeselectIfSameColor(start_object, color) # Iterate through object hierarchy and select if same color

    # Select Xpresso nodes
    if useXpresso == True:
        for s in selection: # Iterate through selected items
            if IdentifyObject(s) == "Xpresso": # If item is Xpresso
                nm = s.GetNodeMaster() # Get node master
                root = nm.GetRoot() # Get node master's root
                nm.AddUndo() # Record undo
                for c in root.GetChildren(): # Iterate through nodes
                    if c[c4d.ID_GVBASE_COLOR] == color: # If node color is same as given color
                        c.DelBit(c4d.BIT_ACTIVE) # Deselect node

    # Select Redshift nodes
    if useRedshift == True:
        materials = doc.GetMaterials() # Get materials
        try: # Try to execute following script
            for m in materials: # Iterate through materials
                if m.GetBit(c4d.BIT_ACTIVE): # If material is selected
                    rsnm = redshift.GetRSMaterialNodeMaster(m) # Get Redshift material node master
                    rsroot = rsnm.GetRoot() # Get node master's root
                    rsnm.AddUndo() # Record undo
                    for r in rsroot.GetChildren(): # Iterate through nodes
                        if r[c4d.ID_GVBASE_COLOR] == color: # If node color is same as given color
                            r.DelBit(c4d.BIT_ACTIVE) # Deselect node
        except: # If something went wrong
            pass # Do nothing

    # Deselect polygons
    if usePolygons == True:
        for s in selection: # Iterate through selected objects
            if s.GetType() == 5100: # If object is a polygon object
                DeselectPolygons(s, id)

    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def GroupUnderNull(color):
    """ Groups selected objects under the null and assigns to the layer """

    def SelectChildren(obj, next):
        doc = c4d.documents.GetActiveDocument() # Get active document
        while obj and obj != next: # Iterate until no children left
            doc.AddUndo(c4d.UNDOTYPE_CHANGE_SMALL, obj) # Record undo
            obj.SetBit(c4d.BIT_ACTIVE) # Select child
            SelectChildren(obj.GetDown(), next) # ;)
            obj = obj.GetNext() # Get next child
        return True

    def Deselect(sel):
        doc = c4d.documents.GetActiveDocument() # Get active document
        for s in sel: # Iterate through selection
            doc.AddUndo(c4d.UNDOTYPE_CHANGE_SMALL, s) # Record undo
            s.DelBit(c4d.BIT_ACTIVE) # Deselect item

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos
    
    c4d.CallCommand(100004772) # Group objects
    null = doc.GetSelection()[0] # Get the null
    null[c4d.ID_BASEOBJECT_USECOLOR] = 2 # 
    null[c4d.NULLOBJECT_DISPLAY] = 0
    #null[c4d.NULLOBJECT_ICONCOL] = 1
    null[c4d.ID_BASEOBJECT_COLOR] = color # Set color
    null[c4d.ID_BASELIST_ICON_COLORIZE_MODE] = 1 # Custom
    null[c4d.ID_BASELIST_ICON_COLOR] = color # Custom Icon Color

    # Select children
    for obj in doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_CHILDREN): # Get active selection
        SelectChildren(obj, obj.GetNext()) # Select children

    selection = doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_CHILDREN) # Get objects including children

    for s in selection: # Iterate through selection
        if not CheckIfLayerExists(color): # If layer does not exist
            CreateLayer(color) # Create a new layer
            AssignToLayer(s, color) # Set layer
        else: # Otherwise
            AssignToLayer(s, color) # Set layer

    Deselect(selection) # Deselect items
    
    doc.AddUndo(c4d.UNDOTYPE_CHANGE_SMALL, null) # Record undo
    null.SetBit(c4d.BIT_ACTIVE) # Select null
    
    doc.EndUndo() # End recording undos
    c4d.EventAdd() # Refresh Cinema 4D
#----------------------------------------------------------------------------------------------------------------------------------------
# M E N U
#----------------------------------------------------------------------------------------------------------------------------------------
def LoadDefaultPalette(self, colorFields):
    """ Loads the default color palette to the Colorise plug-in's color fields """

    try: # Try to execute following script
        path, fn = os.path.split(__file__)
        coloursFile = os.path.join(path, "res", "default.txt")
        coloursArray = []
        if (sys.version_info >= (3, 0)): # If Python 3 version (R23)
            f = open(coloursFile) # Open the file for reading
        else: # If Python 2 version (R21)
            f = open(coloursFile.decode("utf-8"))
        for line in f: # Iterate through every row
            line = line.split(",") # Split by comma
            r = float(line[0]) # Get red value
            g = float(line[1]) # Get green value
            b = float(line[2]) # Get blue value
            coloursArray.append(c4d.Vector(r/255.0,g/255.0,b/255.0)) # Add color to the array
        for i, cf in enumerate(colorFields): # Iterate through colors
            self.SetColorField(cf, coloursArray[i], 1, 1, 0) # Set color to the color field
        f.close() # Close the file
    except: # If something went wrong
        pass # Do nothing
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def LoadFromFile(self, colorFields):
    """ Loads Colorise palette from a file """

    fn = s.LoadDialog(c4d.FILESELECTTYPE_ANYTHING,'Select colour palette file',c4d.FILESELECT_LOAD, '')
    if not fn: return True # If no file path, fuck this!
    if (sys.version_info >= (3, 0)): # If Python 3 version (R23)
        f = open(fn) # Open the file for reading
    else: # If Python 2 version (R21)
        f = open(fn.decode("utf-8"))
    coloursArray = []
    for line in f: # Iterate through every row
        line = line.split(",") # Split by comma
        r = float(line[0]) # Get red value
        g = float(line[1]) # Get green value
        b = float(line[2]) # Get blue value
        coloursArray.append(c4d.Vector(r/255.0,g/255.0,b/255.0)) # Add color to the array
    for i, cf in enumerate(colorFields): # Iterate through colors
        self.SetColorField(cf, coloursArray[i], 1, 1, 0) # Set color to the color field
    f.close() # Close the file
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def SaveToFile(self, colorFields):
    """ Saves Colorise palette to a file """

    lines = [] # Initialize an array for colors
    for cf in colorFields:
        line = ""+str(self.GetColorField(cf)["color"][0]*255)+","+str(self.GetColorField(cf)["color"][1]*255)+","+str(self.GetColorField(cf)["color"][2]*255)+"" # Get color values
        lines.append(line) # Add color values to the array
    lines = "\n".join(lines) # Collect everything to a single string
    fn = s.LoadDialog(c4d.FILESELECTTYPE_ANYTHING,'Save colour palette file',c4d.FILESELECT_SAVE, '') # Save dialog
    if not fn: return True # If no file path, everything is ruined...
    if (sys.version_info >= (3, 0)): # If Python 3 version (R23)
        f = open(fn, 'w') # Open the file for writing
    else: # If Python 2 version (R21)
        f = open(fn.decode("utf-8"), 'w') # Open the file for writing
    f.write(lines) # Write to the file
    f.close() # Close the file
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def SetAsDefault(self, colorFields):
    """ Saves current palette as default """

    lines = [] # Initialize an array for colors
    for cf in colorFields:
        line = ""+str(self.GetColorField(cf)["color"][0]*255)+","+str(self.GetColorField(cf)["color"][1]*255)+","+str(self.GetColorField(cf)["color"][2]*255)+"" # Get color values
        lines.append(line) # Add color values to the array
    lines = "\n".join(lines) # Collect everything to a single string
    fn = os.path.join(path, "res", "default.txt") # File path for the default palette
    if not fn: return True # If no file path, no can do
    if (sys.version_info >= (3, 0)): # If Python 3 version (R23)
        f = open(fn, 'w') # Open the file for writing
    else: # If Python 2 version (R21)
        f = open(fn.decode("utf-8"), 'w') # Open the file for writing
    f.write(lines) # Write to the file
    f.close() # Close the file
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def CreateMaterials(self, colorFields):
    """ Creates materials from current palette's colors """

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos
    for cf in reversed(colorFields): # Iterate through color fields
        mat = c4d.BaseMaterial(c4d.Mmaterial) # Initialize a new material
        mat[c4d.MATERIAL_USE_REFLECTION] = 0 # Disable reflection channel
        mat[c4d.MATERIAL_USE_LUMINANCE] = 0 # Disable luminance channel
        mat[c4d.MATERIAL_USE_COLOR] = 1 # Enable color channel
        mat[c4d.MATERIAL_COLOR_COLOR] = self.GetColorField(cf)["color"] # Set diffuse color
        mat[c4d.MATERIAL_LUMINANCE_COLOR] = self.GetColorField(cf)["color"] # Set luminance color
        mat.Message(c4d.MSG_UPDATE) # Send update message
        mat.Update(True, True) # Update material
        doc.InsertMaterial(mat) # Insert material to document
        doc.AddUndo(c4d.UNDOTYPE_NEW, mat) # Record undo for creating a new material
        c4d.EventAdd() # Refresh Cinema 4D
    doc.EndUndo() # End recording undos
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def CreateSwatches(self, colorFields):
    """ Creates swatches from current palette's colors """

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos
    csd = colorchooser.ColorSwatchData(doc) # Get color swatch data
    csg = colorchooser.ColorSwatchGroup("Colorise", False) # Get color swatch group
    for cf in colorFields: # Iterate through color fields
        c = self.GetColorField(cf)["color"] # Get color from color field
        color = maxon.ColorA64(c[0], c[1], c[2], 1)
        csg.AddColor(color, False, -1)
    csd.InsertGroup(csg, -1, c4d.SWATCH_CATEGORY_DOCUMENT)
    csd.Save(doc) # Save swatches to document
    c4d.EventAdd() # Refresh Cinema 4D
    doc.EndUndo() # End recording undos
    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def PaletteFromSwatches(self, colorFields):
    """ Creates a palette from selected swatches """

    doc = c4d.documents.GetActiveDocument() # Get active document
    csd = colorchooser.ColorSwatchData(doc) # Get color swatch data
    grpCount = csd.GetGroupCount(c4d.SWATCH_CATEGORY_DOCUMENT) # Get swatch group count
    coloursArray = [] # Initialize an array
    replaceColors = [] # Initialize an array
    for i in range(0, 16): # Iterate through
        coloursArray.append(c4d.Vector(0,0,0)) # Fill color array with black colors
    
    for y in range(0, grpCount): # Iterate through swatch groups
        csg = csd.GetGroupAtIndex(y) # Get swatch group
        for y in range(0, csg.GetColorCount()): # Iterate through swatches
            if csg.IsColorSelected(y): # If swatch is selected
                clrs = str(csg.GetColor(y)[0]).split(",") # Get swatch color
                replaceColors.append(c4d.Vector(float(clrs[0]),float(clrs[1]),float(clrs[2]))) # Add color to the array
    if len(replaceColors) < len(coloursArray):
        rangeAmount = len(replaceColors)
    else:
        rangeAmount = len(coloursArray)
    for k in range(0, rangeAmount):
        coloursArray[k] = replaceColors[k]
    for x in range(0, 16): # Iterate through colors
        self.SetColorField(colorFields[x], coloursArray[x], 1, 1, 0) # Set color to the color field

#----------------------------------------------------------------------------------------------------------------------------------------
# L A Y E R S
#----------------------------------------------------------------------------------------------------------------------------------------
def RemoveFromLayer():
    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos
    selection = doc.GetSelection() # Get selected items
    for s in selection: # Iterate through selected items
        if IdentifyObject(s) == "Object" or IdentifyObject(s) == "Null":
            doc.AddUndo(c4d.UNDOTYPE_CHANGE, s) # Record undo
            s.SetLayerObject(None) # Remove layer
    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D

#----------------------------------------------------------------------------------------------------------------------------------------
def CreateLayer(color):
    doc = c4d.documents.GetActiveDocument() # Get active document
    layerRoot = doc.GetLayerObjectRoot() # Get layer object root
    layer = c4d.documents.LayerObject() # Initialize a layer object
    layer.SetName("Layer") # Set layer's name
    layer[c4d.ID_LAYER_COLOR] = color # Set layer's color
    layer.InsertUnder(layerRoot) # Insert layer to layer root
    doc.AddUndo(c4d.UNDOTYPE_NEW, layer) # Record undo for creating a new layer

#----------------------------------------------------------------------------------------------------------------------------------------
def CollectLayers():
    """ Collects all layers """

    def IterateLayers(op):
        layerList = [] # Initialize an array for collecting layers
        if op is None: # If there is no layer
            return # This is over
        while op: # While there is an item
            layerList.append(op) # Add layer to layer list
            op = GetNextItem(op) # Get next layer
        return layerList # Return layers

    doc = c4d.documents.GetActiveDocument() # Get active document
    layerRoot = doc.GetLayerObjectRoot() # Get layer object root
    layers = layerRoot.GetChildren() # Get layers
    if layers == []: # Check if there is no any layer
        return None # Return none
    else: # If there is any layer
        startLayer = layers[0] # Get start layer for iterating through all layers
        return IterateLayers(startLayer) # Return collection of all layers

#----------------------------------------------------------------------------------------------------------------------------------------
def CheckIfLayerExists(color):
    layerExists = False # Initialize a variable for storing status of layer exitance
    allLayers = CollectLayers() # Collect all layers
    if allLayers == None: # If there is no layers
        return layerExists # Layer does not exist
    else: # If there is layers
        for layer in allLayers: # Loop through layers
            if layer[c4d.ID_LAYER_COLOR] == color: # If layer with same color is found
                layerExists = True # Layer does exist
        return layerExists # Layer does not exist

#----------------------------------------------------------------------------------------------------------------------------------------
def SetLayer(color):
    """ Set object's layer to given color """

    doc = c4d.documents.GetActiveDocument() # Get active document
    doc.StartUndo() # Start recording undos
    selection = doc.GetSelection() # Get active selection
    for s in selection: # Iterate through selected items
        if IdentifyObject(s) == "Object" or IdentifyObject(s) == "Null":
            if not CheckIfLayerExists(color): # If layer does not exist
                CreateLayer(color) # Create layer with given color
                AssignToLayer(s, color) # Set layer
            else: # Layer exists
                AssignToLayer(s, color) # Set layer

    doc.EndUndo() # Stop recording undos
    c4d.EventAdd() # Refresh Cinema 4D
#----------------------------------------------------------------------------------------------------------------------------------------
"""




And then something very different!





"""
#----------------------------------------------------------------------------------------------------------------------------------------
def LoadOptions(self):
    """ Loads Colorise's options """

    global useIcons
    global useObjects
    global useLights
    global useXpresso
    global useDoodle
    global useRedshift
    global forceNulls
    global usePolygons

    self.SetBool(2002, useIcons) # Colorise Icons
    self.SetBool(2003, useObjects) # Colorise Objects
    self.SetBool(2012, useLights) # Colorise Lights
    self.SetBool(2004, useXpresso) # Colorise Xpresso
    self.SetBool(2005, useDoodle) # Colorise Doodle Tool
    self.SetBool(2006, useRedshift) # Colorise Redshift
    self.SetBool(2011, forceNulls) # Force Nulls
    self.SetBool(2013, usePolygons) # Create Polygon selections


    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
def SaveOptions(self):
    """ Saves Colorise's options """

    global useIcons
    global useObjects
    global useLights
    global useXpresso
    global useDoodle
    global useRedshift
    global forceNulls
    global usePolygons

    path, fn = os.path.split(__file__)
    optionsFile = os.path.join(path, "res", "options.txt")
    if (sys.version_info >= (3, 0)): # If Python 3 version (R23)
        f = open(optionsFile, 'w') # Open the file for writing
    else: # If Python 2 version (R21)
        f = open(optionsFile.decode("utf-8"), 'w') # Open the file for writing
    settings = [ str(self.GetBool(2002)), str(self.GetBool(2003)), str(self.GetBool(2004)),
                 str(self.GetBool(2005)), str(self.GetBool(2006)), str(self.GetBool(2011)),
                 str(self.GetBool(2012)), str(self.GetBool(2013))] # Get checkbox values
    settings = "\n".join(settings) # Create a string from an array
    f.write(settings) # Write settings to the file
    f.close() # Close the file

    useIcons   = self.GetBool(2002)
    useObjects = self.GetBool(2003)
    useLights  = self.GetBool(2012)
    useXpresso = self.GetBool(2004)
    useDoodle  = self.GetBool(2005)
    useRedshift= self.GetBool(2006)
    forceNulls = self.GetBool(2011)
    usePolygons= self.GetBool(2013)

    return True # Everything is fine

#----------------------------------------------------------------------------------------------------------------------------------------
# Options GUI
#----------------------------------------------------------------------------------------------------------------------------------------
class OptionsDialog(gui.GeDialog):
    """ Creates graphical user interface of options dialog  """

    #def __init__(self, Value):
        #self.Value = Value
        
    def CreateLayout(self):
        self.SetTitle("Options")

        self.GroupBegin(2000, c4d.BFH_CENTER, 1, 3) # Begin main group

        self.GroupBegin(2001, c4d.BFH_LEFT, cols=1, rows=4, title="Enable", groupflags=0, initw=0, inith=0) # Begin checkbox group
        self.GroupBorder(c4d.BORDER_IN)
        self.GroupSpace(10, 10)
        self.GroupBorderSpace(left=5, top=5, right=5, bottom=5)
        self.AddCheckbox(2002, flags=c4d.BFH_LEFT, initw=0, inith=0, name="Use Icons")
        self.AddCheckbox(2003, flags=c4d.BFH_LEFT, initw=0, inith=0, name="Use Objects")
        self.AddCheckbox(2012, flags=c4d.BFH_LEFT, initw=0, inith=0, name="Use Lights")
        self.AddCheckbox(2004, flags=c4d.BFH_LEFT, initw=0, inith=0, name="Use Xpresso")
        self.AddCheckbox(2005, flags=c4d.BFH_LEFT, initw=0, inith=0, name="Use Doodle Tool")
        self.AddCheckbox(2006, flags=c4d.BFH_LEFT, initw=0, inith=0, name="Use Redshift")
        self.AddCheckbox(2013, flags=c4d.BFH_LEFT, initw=0, inith=0, name="Use Polygons")
        self.GroupEnd() # End checkbox group

        self.GroupBegin(2010, c4d.BFH_LEFT, cols=1, rows=4, title="Override", groupflags=0, initw=0, inith=0) # Begin override group
        self.GroupBorder(c4d.BORDER_IN)
        self.GroupSpace(10, 10)
        self.GroupBorderSpace(left=5, top=5, right=5, bottom=5)
        self.AddCheckbox(2011, flags=c4d.BFH_LEFT, initw=0, inith=0, name="Nulls Icons")
        self.GroupEnd() # End override group


        self.GroupBegin(2007, c4d.BFH_CENTER, 2, 1) # Begin button group
        self.GroupSpace(10, 10)
        self.AddButton(2008, c4d.BFH_LEFT, initw=0, inith=0, name="Save")
        self.AddButton(2009, c4d.BFH_RIGHT, initw=0, inith=0, name="Cancel")
        self.GroupEnd() # End button group

        self.GroupEnd() # End main group

        LoadOptions(self) # Load options

        return True

    def Command(self, id, msg):
        # Save button is pressed
        if (id == 2008):
            SaveOptions(self) # Save options
            self.Close() # Close the dialog
            return True

        # Cancel button is pressed
        if (id == 2009):
            self.Close() # Close the dialog
            return True

        return True # Everything is fine
#----------------------------------------------------------------------------------------------------------------------------------------
# Main Plug-in GUI
#----------------------------------------------------------------------------------------------------------------------------------------
class ColoriseDialog(gui.GeDialog):
    """ Creates the main graphical user interface of the plug-in  """

    def CreateLayout(self):
        self.SetTitle("Colorise") # Set dialog title

        # Submenu "File"
        self.MenuSubBegin("File")
        self.MenuAddString(11050,"Options")
        self.MenuAddSeparator() # Add separator line
        self.MenuAddString(11052,"Load from file")
        self.MenuAddString(11053,"Save to file")
        self.MenuAddSeparator() # Add separator line
        self.MenuAddString(11051,"Reset palette")
        self.MenuAddString(11054,"Set as default")
        self.MenuAddSeparator() # Add separator line
        self.MenuAddString(11056,"Palette from swatches")
        self.MenuAddString(11057,"Create swatches")
        self.MenuAddString(11058,"Create materials")
        self.MenuSubEnd() # File submenu finished

        # Submenu "Presets"
        self.MenuSubBegin("Presets")
        presetBase = 11100 # ID base for preset items
        path, fn = os.path.split(__file__)
        i = 0 # Initialize an iteration variable
        try: # Try to execute following script
            for file in os.listdir(path+"/presets"): # Iterate through files in presets folder
                if file.endswith(".txt"): # If file extension is txt
                    presetPath = os.path.join(path+"/presets", file) # Get preset path
                    presetName = os.path.splitext(file)[0] # Get preset name
                    presetId = presetBase+i # Current preset ID
                    self.MenuAddString(presetId, presetName) # Add preset to the submenu
                    presets[presetId] = presetPath # Add preset to the presets dictionary
                    i += 1 # You know right...
        except: # If something went wrong
            pass # Do nothing
        self.MenuSubEnd() # File submenu finished
        self.MenuFinished() # Dialog menu finished

        # Generate GUI content
        self.GroupBegin(1000, c4d.BFH_LEFT | c4d.BFV_TOP, 1, 17, c4d.BFV_GRIDGROUP_EQUALCOLS) # Begin main Group

        colorBase = 11000 # ID base for color fields
        buttonBase = 11020 # ID base for buttons

        for x in range(0, 16): # Generate buttons and color fields        
            colorId = colorBase + x # Current color field ID
            buttonId = buttonBase + x # Current button ID
            self.GroupBegin((1001+x), c4d.BFH_CENTER | c4d.BFV_TOP, 2, 1, c4d.BFV_GRIDGROUP_EQUALCOLS) # Begin button/color field group
            self.AddButton(buttonId, c4d.BFH_CENTER | c4d.BFV_TOP, initw=14, inith=10, name=" ") # Add a button
            self.AddColorField(colorId, c4d.BFH_CENTER | c4d.BFV_TOP, 50) # Add a color field
            self.GroupEnd() # End button/color field group

        # Default button
        self.GroupBegin(1018, c4d.BFH_CENTER | c4d.BFV_TOP, 2, 1, c4d.BFV_GRIDGROUP_EQUALCOLS) # Begin default button group
        self.AddButton(11036, c4d.BFH_CENTER | c4d.BFV_TOP, initw=50, inith=14, name="X") # Add a button
        self.GroupEnd() # End default button group
        
        self.GroupEnd() # End main group

        LoadDefaultPalette(self, colorFields) # Loads default template palette from a file
        LoadSettings() # Load Colorise's settings

        return True # Everything is fine

    # Handle commands, pressing menu items, muttons, etc.

    def Command(self, id, msg):
        colorBase = 11000 # ID base for color fields

        # Check keyboard modifiers
        doc = c4d.documents.GetActiveDocument() # Get active document
        bc = c4d.BaseContainer() # Initialize a base container
        keyMod = "None" # Initialize a keyboard modifier status
        # Button is pressed
        if c4d.gui.GetInputState(c4d.BFM_INPUT_KEYBOARD,c4d.BFM_INPUT_CHANNEL,bc):
            if bc[c4d.BFM_INPUT_QUALIFIER] & c4d.QSHIFT:
                if bc[c4d.BFM_INPUT_QUALIFIER] & c4d.QCTRL: # Shift + Ctrl
                    if bc[c4d.BFM_INPUT_QUALIFIER] & c4d.QALT: # Shift + Ctrl + Alt
                        keyMod = 'Alt+Ctrl+Shift'
                    else: # Ctrl + Shift
                        keyMod = 'Ctrl+Shift'
                elif bc[c4d.BFM_INPUT_QUALIFIER] & c4d.QALT: # Shift + Alt
                    keyMod = 'Alt+Shift'
                else: # Shift
                    keyMod = 'Shift'
            elif bc[c4d.BFM_INPUT_QUALIFIER] & c4d.QCTRL:
                if bc[c4d.BFM_INPUT_QUALIFIER] & c4d.QALT: # Ctrl + Alt
                    keyMod = 'Alt+Ctrl'
                else: # Ctrl
                    keyMod = 'Ctrl'
            elif bc[c4d.BFM_INPUT_QUALIFIER] & c4d.QALT: # Alt
                keyMod = 'Alt'
            else: # No keyboard modifiers used
                keyMod = 'None'

        #--------------------------------------------------------------------------------------------------------------------------------
        # B U T T O N
        #--------------------------------------------------------------------------------------------------------------------------------
        # Color button is pressed
        if (id in colorsDict):
            color = self.GetColorField(colorsDict[id])["color"] # Get color
            if keyMod == 'None': # No keyboard modifiers used
                SetColor(color, id) # Run set color function

            elif keyMod == 'Alt': # Alt pressed
                SelectItem(color, False, id) # Select items with given color

            elif keyMod == 'Ctrl': # Ctrl pressed
                GroupUnderNull(color) # Group selected items under the null

            elif keyMod == 'Shift': # Alt pressed
                SetLayer(color) # Set selected items layers to given color

            elif keyMod == 'Alt+Shift': # Alt pressed
                SelectItem(color, True, id) # Add items with given color to current selection

            elif keyMod == 'Alt+Ctrl': # Alt + Ctrl pressed
                RemoveFromSelection(color, id) # Removes items from selection that are colored as given color

        # Default color is pressed
        if (id == 11036):
            if keyMod == 'None': # No keyboard modifiers used
                ResetColor() # Run reset color function

            elif keyMod == 'Alt': # Alt pressed
                SequenceColors(self) # Sequences colors to selected items

            elif keyMod == 'Ctrl': # Ctrl pressed
                RandomColors(self) # Random colors from palette to selected items

            elif keyMod == 'Shift': # Shift pressed
                RemoveFromLayer() # Remove selected items from given color layer


        #--------------------------------------------------------------------------------------------------------------------------------
        # M E N U
        #--------------------------------------------------------------------------------------------------------------------------------
        # 'Options' menu item is pressed
        if (id == 11050):
            self.subDialog = OptionsDialog() # Create options dialog object
            self.subDialog.Open(dlgtype=c4d.DLG_TYPE_MODAL, xpos=-1, ypos=-1, defaultw=0, defaulth=0) # Open options dialog

        # 'Load from file' menu item is pressed
        if (id == 11052):
            LoadFromFile(self, colorFields)

        # 'Save to file' menu item is pressed
        if (id == 11053):
            SaveToFile(self, colorFields)

        # 'Reset Palette' menu item is pressed
        if (id == 11051):
            LoadDefaultPalette(self, colorFields)

        # 'Set as default' menu item is pressed
        if (id == 11054):
            SetAsDefault(self, colorFields)

        # 'Palette from swatches' menu item is pressed
        if (id == 11056):
            PaletteFromSwatches(self, colorFields)

        # 'Create swatches' menu item is pressed
        if (id == 11057):
            CreateSwatches(self, colorFields)

        # 'Create materials' menu item is pressed
        if (id == 11058):
            CreateMaterials(self, colorFields)

        # Loading a preset
        if (id in presets): # If preset is called
            path, fn = os.path.split(__file__) # Get path
            fn = os.path.join(path, "presets", presets[id]) # Get preset's file path
            if not fn: return True # If file path does not exists, abandon the ship
            if (sys.version_info >= (3, 0)): # If Python 3 version (R23)
                f = open(fn) # Open the file for reading
            else: # If Python 2 version (R21)
                f = open(fn.decode("utf-8"))
            coloursArray = [] # Initialize an array for colors
            for line in f: # Iterate through rows
                line = line.split(",") # Split by comma
                r = float(line[0]) # Get red value
                g = float(line[1]) # Get green value
                b = float(line[2]) # Get blue value
                coloursArray.append(c4d.Vector(r/255.0,g/255.0,b/255.0)) # Add color to the array
            for x in range(0, 16): # Iterate through colors
                self.SetColorField(colorFields[x], coloursArray[x], 1, 1, 0) # Set color to the color field
            f.close() # Close the file
            return True

        return True # Everything is fine
#----------------------------------------------------------------------------------------------------------------------------------------
# Plug-in class
#----------------------------------------------------------------------------------------------------------------------------------------
class Colorise(plugins.CommandData):
    dialog = None
    def Execute(self, doc):
        if self.dialog is None:
            self.dialog = ColoriseDialog()
        return self.dialog.Open(dlgtype=c4d.DLG_TYPE_ASYNC, pluginid=PLUGIN_ID, defaultw=400, defaulth=50, xpos=-1, ypos=-1)
        
        
    def RestoreLayout(self, sec_ref):
        if self.dialog is None:
            self.dialog = ColoriseDialog()
        return self.dialog.Restore(pluginid=PLUGIN_ID, secret=sec_ref)

#----------------------------------------------------------------------------------------------------------------------------------------
# Main
#----------------------------------------------------------------------------------------------------------------------------------------
if __name__ == "__main__":
    dir, file = os.path.split(__file__)
    path, fn = os.path.split(__file__)
    bmp = bitmaps.BaseBitmap()
    bmp.InitWith(os.path.join(path, "res", "icon.tif"))
    plugins.RegisterCommandPlugin(PLUGIN_ID, "Colorise",0, bmp, "Colorise", Colorise())